# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

import datetime

# ISO 8601 extended time format without microseconds
_ISO8601_TIME_FORMAT: str = '%Y-%m-%dT%H:%M:%S'


def isotime(at: datetime.datetime) -> str:
    """Stringify time in ISO 8601 format."""
    st = at.strftime(_ISO8601_TIME_FORMAT)
    tz: str | None = at.tzinfo.tzname(None) if at.tzinfo else 'UTC'
    # Need to handle either iso8601 or python UTC format
    if tz is None or tz in ['UTC', 'UTC+00:00']:
        st += 'Z'
    else:
        st += tz
    return st
